theory CanonicalEq
imports Nominal Name HOAS
begin

text {* Canonical forms, existence, and uniqueness modulo beta-eta equivalence *}

inductive of_atm :: "hoas_sig \<Rightarrow> hoas_ctx \<Rightarrow> hoas \<Rightarrow> hoas_ty \<Rightarrow> bool"
  ("_,_ \<turnstile> _ \<down>: _" [60,60,60,60] 60)
  and   of_can :: "hoas_sig \<Rightarrow> hoas_ctx \<Rightarrow> hoas \<Rightarrow> hoas_ty \<Rightarrow> bool"
  ("_,_ \<turnstile> _ \<Up>: _" [60,60,60,60] 60)
  where
  of_VAR[intro] : "(v,ty) \<in> set ctx \<Longrightarrow> of_atm sig ctx (VAR v) ty"
| of_APP[intro] : "\<lbrakk>of_atm sig ctx t1 (FUN ty1 ty2);
                    of_can sig ctx t2 ty1\<rbrakk> \<Longrightarrow> 
                    of_atm sig ctx (APP t1 t2) ty2"
| of_CST[intro] : "(c,ty) \<in> set sig \<Longrightarrow> of_atm sig ctx (CST c) ty"
| of_INC[intro] : "of_atm sig ctx t (BASE ty) \<Longrightarrow> 
                   of_can sig ctx t (BASE ty)"
| of_LAM[intro] : "\<lbrakk>x \<sharp> ctx;of_can sig ((x,ty1)#ctx) t1 ty2\<rbrakk> \<Longrightarrow> 
                    of_can sig ctx (LAM x t1) (ty1 ~> ty2)"

equivariance of_atm

nominal_inductive of_atm
  apply(auto simp add: abs_fresh )
  done


inductive_cases of_atm_can_inv[elim!]:
  "of_atm sig ctx (VAR v) ty"
  "of_atm sig ctx (APP t1 t2) ty2"
  "of_atm sig ctx (CST c) ty"  
  "of_can sig ctx t (BASE ty)"
  "of_can sig ctx (LAM x t1) (ty1 ~> ty2)"

text {* Proof that canonical/atomic forms are ordinarily well-formed *}

theorem of_wf : "of_atm sig ctx t ty \<Longrightarrow> wf sig ctx t ty" 
                "of_can sig ctx t ty \<Longrightarrow> wf sig ctx t ty"
  by(nominal_induct sig ctx t ty and sig ctx t ty 
    rule: of_atm_of_can.strong_inducts)
  (auto)




inductive to_atm :: "hoas_sig \<Rightarrow> hoas_ctx \<Rightarrow> hoas \<Rightarrow> hoas \<Rightarrow> hoas_ty \<Rightarrow> bool"    
  ("_,_ \<turnstile> _ \<down> _ : _" [60,60,60,60,60] 60)
and to_can :: "hoas_sig \<Rightarrow> hoas_ctx \<Rightarrow> hoas \<Rightarrow> hoas \<Rightarrow> hoas_ty \<Rightarrow> bool"  
  ("_,_ \<turnstile> _ \<Up> _ : _" [60,60,60,60,60] 60)
  
where
  to_atm_var: "(x,\<tau>) \<in> set \<Gamma> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> VAR x \<down> VAR x : \<tau>"
| to_atm_cst: "(c,\<tau>) \<in> set \<Sigma> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> CST c \<down> CST c : \<tau>"
| to_atm_app: "\<lbrakk>\<Sigma>,\<Gamma> \<turnstile> M \<down> M' : \<tau>' ~> \<tau> ; \<Sigma>,\<Gamma> \<turnstile> N \<Up> N' : \<tau>'\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> APP M N \<down> APP M' N' : \<tau>"
| to_can_atm: "\<Sigma>,\<Gamma> \<turnstile> M \<down> N : BASE a \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> M \<Up> N : BASE a"
| to_can_fun: "\<lbrakk>\<Sigma>,(x,\<tau>)#\<Gamma> \<turnstile> APP M (VAR x) \<Up> N : \<tau>';x \<sharp> (\<Gamma>,M) \<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> M \<Up> LAM x N : \<tau> ~> \<tau>'"
| to_can_whr: "\<lbrakk>M \<leadsto> M'; \<Sigma>,\<Gamma> \<turnstile> M' \<Up> N : BASE a\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> M \<Up> N : BASE a"

equivariance to_can

nominal_inductive to_can
  by(auto simp add:fresh_string abs_fresh subst_fresh )



lemma to_fresh:
  fixes x::var
  shows "\<lbrakk>\<Sigma>,\<Gamma> \<turnstile> M \<down> K : \<tau>; x \<sharp> M\<rbrakk> \<Longrightarrow> x \<sharp> K"
  and   "\<lbrakk>\<Sigma>,\<Gamma> \<turnstile> M \<Up> K : \<tau>; x \<sharp> M\<rbrakk> \<Longrightarrow> x \<sharp> K"
  apply(nominal_induct \<Sigma> \<Gamma> M K \<tau> and \<Sigma> \<Gamma> M K \<tau> avoiding: x 
    rule:to_atm_to_can.strong_inducts)
  apply(auto simp add:whr_fresh abs_fresh fresh_atm)
  done


theorem soundness: 
  "\<Sigma>,\<Gamma> \<turnstile> M \<down> N : \<tau> \<Longrightarrow> of_atm \<Sigma> \<Gamma> N \<tau>"
  "\<Sigma>,\<Gamma> \<turnstile> M \<Up> N : \<tau> \<Longrightarrow> of_can \<Sigma> \<Gamma> N \<tau>"
  apply(nominal_induct \<Sigma> \<Gamma> M N \<tau> and \<Sigma> \<Gamma> M N \<tau> rule:to_atm_to_can.strong_inducts)
  by(auto)



lemma to_can_whr_strong_: 
  "\<lbrakk>\<Sigma>,\<Gamma> \<turnstile> M' \<Up> N : \<tau>; M \<leadsto> M'\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> M \<Up> N : \<tau>"
  apply(induct \<tau> arbitrary: M M' N \<Gamma> rule:hoas_ty.inducts)
  apply(erule to_can_whr,blast)
  apply(rule_tac x="(\<Gamma>,M,M',N)" in  exists_fresh,finite_guess)
  apply(erule_tac x="c" in to_can.strong_cases,simp_all add:hoas_ty.inject,clarify)
  apply(subgoal_tac "APP M (VAR c) \<leadsto> APP Ma (VAR c)")
  prefer 2 apply(erule whr_app)
  apply(rule to_can_fun)
  apply(blast)
  apply(simp add:fresh_prod) 
  done

lemma to_can_whr_strong: 
  "\<lbrakk>M \<leadsto> M'; \<Sigma>,\<Gamma> \<turnstile> M' \<Up> N : \<tau>\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> M \<Up> N : \<tau>"
  by(auto intro:to_can_whr_strong_)

lemma canonical_equals_can:
  "of_atm \<Sigma> \<Gamma> M \<tau> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> M \<down> M : \<tau>"
  "of_can \<Sigma> \<Gamma> M \<tau> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> M \<Up> M : \<tau>"
  apply(nominal_induct \<Sigma> \<Gamma> M \<tau> and \<Sigma> \<Gamma> M \<tau>  rule: of_atm_of_can.strong_inducts)
  apply(auto intro:to_atm_to_can.intros)
  apply(rule to_can_fun,simp_all add:abs_fresh fresh_prod)
  apply(rule to_can_whr_strong)
  apply(rule whr_beta_strong, simp add:subst_id)
  done

lemma to_soundness:
  "\<Sigma>,\<Gamma> \<turnstile> M \<down> N : \<tau> \<Longrightarrow>  \<Sigma>,\<Gamma> \<turnstile> N : \<tau>"
  "\<Sigma>,\<Gamma> \<turnstile> M \<Up> N : \<tau> \<Longrightarrow>  \<Sigma>,\<Gamma> \<turnstile> N : \<tau>"
  apply(auto intro: of_wf soundness)
  done


theorem equivalence:
  "\<Sigma>,\<Gamma> \<turnstile> M \<down> N : \<tau> \<Longrightarrow> valid_sig \<Sigma> \<Longrightarrow> valid_ctx \<Gamma> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> M : \<tau>' \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> M \<approx> N : \<tau> \<and> \<tau>=\<tau>'"
  "\<Sigma>,\<Gamma> \<turnstile> M \<Up> N : \<tau> \<Longrightarrow> valid_sig \<Sigma> \<Longrightarrow> valid_ctx \<Gamma> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> M : \<tau> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> M \<approx> N : \<tau>"
  apply(nominal_induct \<Sigma> \<Gamma> M N \<tau> and \<Sigma> \<Gamma> M N \<tau> avoiding: \<tau>' rule:to_atm_to_can.strong_inducts)
  (* case 1 *)
  apply(erule wf.cases, simp_all (no_asm_use) add:hoas.inject,clarify)
  apply(simp add:ctx_det)
  apply(clarify)
(* case 2 *)
  apply(erule wf.cases, simp_all (no_asm_use) add:hoas.inject,clarify)
  apply(simp add:sig_det)
  apply(clarify)
(* case 3 *)
  apply(erule wf.cases,simp_all (no_asm_use) add:hoas.inject,clarify)
  apply(drule_tac x="ty1 ~> ty2" in meta_spec) 
  apply(simp add:hoas_ty.inject,clarify)
  apply(blast intro:wf_equiv_app)
(* case 4 *)
  apply(blast)
(* case 5 *)
  apply(subgoal_tac "valid_ctx ((x, \<tau>) # \<Gamma>)")
  prefer 2 apply(blast)
  apply(subgoal_tac "\<Sigma>,(x, \<tau>) # \<Gamma> \<turnstile> APP M (VAR x) : \<tau>'")
  prefer 2 apply(rule wf_APP, rule weakening,auto)
  apply(rule wf_equiv_trans)
  apply(erule_tac x="x" in wf_equiv_eta)
  apply(simp add:fresh_prod)
  apply(erule wf_equiv_lam)
  apply(simp)
(* case 6 *)
  apply(frule whr_subject_reduction,auto)
  apply(frule whr_subject_reduction_ty,auto)
  apply(erule wf_equiv_trans,auto)
  done


lemma to_atm_neut: 
  "\<Sigma>,\<Gamma> \<turnstile> M \<down> N : \<tau> \<Longrightarrow> neut M"
  apply(nominal_induct \<Sigma> \<Gamma> M N \<tau>  rule:to_atm_to_can.strong_inducts(1))
  apply(auto)
  done

lemma to_atm_whnf:   "\<Sigma>,\<Gamma> \<turnstile> M \<down> N : \<tau> \<Longrightarrow> whnf M"
  apply(drule to_atm_neut, auto)
  done


lemma to_atm_whr_impos: "\<Sigma>,\<Gamma> \<turnstile> M \<down> N : \<tau> \<Longrightarrow> M \<leadsto> M' \<Longrightarrow> False"
  apply(drule to_atm_whnf, auto intro: whr_whnf_impos)
  done

lemma to_can_whnf_to_atm:
  "\<lbrakk>\<Sigma>,\<Gamma> \<turnstile> M \<Up> N: BASE a; whnf(M)\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> M \<down> N: BASE a"
  apply(erule to_can.cases, 
    simp_all add:hoas_ty.inject, 
    auto dest: whr_whnf_impos)
  done

lemma to_whnf:
  "\<Sigma>,\<Gamma> \<turnstile> M \<down> K : \<tau> \<Longrightarrow> neut(K)"
  "\<Sigma>,\<Gamma> \<turnstile> M \<Up> K : \<tau> \<Longrightarrow> whnf(K)"
  apply(nominal_induct \<Sigma> \<Gamma> M K \<tau> and \<Sigma> \<Gamma> M K \<tau>  rule:to_atm_to_can.strong_inducts)
  by(auto)

lemma to_can_whexp_: 
  "\<Sigma>,\<Gamma> \<turnstile> M \<Up> N : \<tau> \<Longrightarrow> M \<leadsto> M' \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> M' \<Up> N : \<tau>"
  apply(nominal_induct \<Sigma> \<Gamma> M N \<tau> avoiding: M' rule:to_atm_to_can.strong_inducts(2))
  apply(auto)
  apply(drule to_atm_whr_impos,auto)
  apply(rule to_can_fun,blast, simp add: fresh_prod)
  apply(drule whr_det,blast,simp)
  done


lemma to_can_whexp: "M \<leadsto> M' \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> M \<Up> N : \<tau> \<Longrightarrow>  \<Sigma>,\<Gamma> \<turnstile> M' \<Up> N : \<tau>"
  by(auto intro:to_can_whexp_)

(* TODO: Can we prove this more directly now using strong_cases? *)


lemma to_can_fn_lam:
  "\<Sigma>,\<Gamma> \<turnstile> M \<down> N : \<tau> \<Longrightarrow> True"
  "\<Sigma>,\<Gamma> \<turnstile> M \<Up> N : \<tau> ~> \<tau>' \<Longrightarrow> x \<sharp> (\<Gamma>,M,N) \<Longrightarrow> (\<exists>K. N = LAM x K \<and> \<Sigma>,(x,\<tau>)#\<Gamma> \<turnstile> APP M (VAR x) \<Up> K : \<tau>')"
  apply(nominal_induct \<Sigma> \<Gamma> M N \<tau> and \<Sigma> \<Gamma> M N \<tau>\<equiv>"\<tau> ~> \<tau>'" avoiding: x \<tau> \<tau>' rule:to_atm_to_can.strong_inducts)
  apply(auto simp add:hoas_ty.inject fresh_prod abs_fresh)
  apply(rule_tac x="[(x,xa)]\<bullet> N" in exI)
  apply(simp add:alpha hoas.inject fresh_atm)
  apply(perm_simp add: fresh_left)
  apply(drule_tac pi="[(x,xa)]" in to_can.eqvt)
  apply(perm_simp)
  done

lemma to_weakening:
  shows "\<Sigma>,\<Gamma> \<turnstile> M \<down> N : \<tau> \<Longrightarrow> \<Gamma>' extends \<Gamma> ==> \<Sigma>,\<Gamma>' \<turnstile> M \<down> N : \<tau>"
  and "\<Sigma>,\<Gamma> \<turnstile> M \<Up> N : \<tau> \<Longrightarrow> \<Gamma>' extends \<Gamma> ==> \<Sigma>,\<Gamma>' \<turnstile> M \<Up> N : \<tau>"
  apply(nominal_induct \<Sigma> \<Gamma> M N \<tau> and \<Sigma> \<Gamma> M N \<tau> 
    avoiding: \<Gamma>'  
    rule: to_atm_to_can.strong_inducts)
  apply(simp)
  apply(blast intro: to_atm_to_can.intros)
  apply(blast intro:to_atm_to_can.intros)+
  apply(rule to_atm_to_can.intros, simp_all add:fresh_prod)
  apply(drule_tac x="((x,\<tau>)#\<Gamma>')" in meta_spec)
  apply(simp add:valid_ctx_cons)
  apply(blast)
  apply(blast intro: to_atm_to_can.intros)
  done
 

theorem to_determinacy: 
  shows   "\<Sigma>,\<Gamma> \<turnstile> M \<down> N : \<tau> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> M \<down> N' : \<tau>' \<Longrightarrow> valid_ctx \<Gamma> \<Longrightarrow> valid_sig \<Sigma> \<Longrightarrow> N = N' \<and> \<tau> = \<tau>'"
  and "\<Sigma>,\<Gamma> \<turnstile> M \<Up> N : \<tau> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> M \<Up> N' : \<tau> \<Longrightarrow> valid_ctx \<Gamma> \<Longrightarrow> valid_sig \<Sigma> \<Longrightarrow> N = N'"
  apply(nominal_induct \<Sigma> \<Gamma> M N \<tau> and \<Sigma> \<Gamma> M N \<tau> avoiding: N' \<tau>' rule:to_atm_to_can.strong_inducts)
  (* case 1 *)
  apply(erule to_atm.cases)
  apply(simp_all (no_asm_use) add:hoas.inject)+
  apply(simp add:hoas.inject)
  apply(rule ctx_det,simp,simp,simp)
    (* case 2 *)
  apply(erule to_atm.cases)
  apply(simp_all (no_asm_use) add:hoas.inject)
  apply(simp add:hoas.inject)
  apply(rule sig_det,simp,simp,simp)
    (* case 3 *)
  apply(erule to_atm.cases) back
  apply(simp_all (no_asm_use) add:hoas.inject)
  apply(clarify, simp (no_asm_use) add:hoas.inject)
  apply(subgoal_tac "M' = M'a \<and> \<tau>' ~> \<tau> = \<tau>'b ~> \<tau>''")
  prefer 2 apply(blast)
  apply(simp (no_asm_use) add:hoas_ty.inject,clarify,blast)
  

  (* case 1 *)
  apply(erule to_can.cases)
  apply(simp_all (no_asm_use) add:hoas_ty.inject)
  apply(blast)
  apply(drule to_atm_whr_impos,simp,simp)

    (* case 2 *)
  apply(drule_tac x="x" in to_can_fn_lam(2))
  apply(simp add:fresh_prod)
  apply(clarify)
  apply(blast)

    (* case 3 *)
  apply(drule to_can_whexp)
  apply(blast)
  apply(blast)
  done
  





function 
  logrel ::  "hoas_sig \<Rightarrow> hoas_ctx \<Rightarrow> hoas \<Rightarrow> hoas \<Rightarrow> hoas_ty \<Rightarrow> bool" ("_,_ \<turnstile> _ = _ \<in> \<lbrakk>_\<rbrakk>" [60,60,60,60] 60) 
where    
   "\<Sigma>,\<Delta> \<turnstile> M = M' \<in> \<lbrakk>BASE a\<rbrakk> = (\<exists> N. \<Sigma>,\<Delta> \<turnstile> M \<Up> N : BASE a \<and> \<Sigma>,\<Delta> \<turnstile> M' \<Up> N : BASE a )"
 | "\<Sigma>,\<Delta> \<turnstile> M = M' \<in> \<lbrakk>\<tau>1 ~> \<tau>2\<rbrakk> = 
          (\<forall>\<Delta>' M1 M1'. \<Delta>' extends \<Delta> \<longrightarrow> \<Sigma>,\<Delta>' \<turnstile> M1 = M1' \<in> \<lbrakk>\<tau>1\<rbrakk> \<longrightarrow> \<Sigma>,\<Delta>' \<turnstile> APP M M1 = APP M' M1' \<in> \<lbrakk>\<tau>2\<rbrakk>)"
  apply (auto simp add:  hoas_ty.inject)
  apply (subgoal_tac "(\<exists>\<tau>1 \<tau>2. b=\<tau>1 ~> \<tau>2) \<or> (\<exists> c. b=BASE  c)")
  apply (force)
  apply(simp add:hoas_ty_cases)
  done

termination
  apply(relation "measure (\<lambda>(_,_,_,_,T). size T)")
  by (auto)


inductive logrel_subst :: "hoas_sig \<Rightarrow> hoas_ctx \<Rightarrow> hoas_subst \<Rightarrow> hoas_subst \<Rightarrow>hoas_ctx \<Rightarrow> bool" ("_,_ \<turnstile> _ = _ \<in> \<lbrakk>_\<rbrakk>" [60,60,60,60] 60) 
where
  logrel_subst_nil : "\<Sigma>,\<Gamma> \<turnstile> [] = [] \<in> \<lbrakk>[]\<rbrakk>"
| logrel_subst_cons : "\<lbrakk>\<Sigma>,\<Gamma> \<turnstile> \<theta> = \<sigma> \<in> \<lbrakk>\<Delta>\<rbrakk>; \<Sigma>,\<Gamma> \<turnstile> M = N \<in> \<lbrakk>\<tau>\<rbrakk>; x\<sharp>\<Delta>\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> (x,M)#\<theta> = (x,N)#\<sigma> \<in> \<lbrakk>(x,\<tau>)#\<Delta>\<rbrakk>"

inductive_cases logrel_subst_inv:
  "\<Sigma>,\<Gamma> \<turnstile> \<theta> = \<sigma> \<in> \<lbrakk>[]\<rbrakk>"
  "\<Sigma>,\<Gamma> \<turnstile> [] = \<sigma> \<in> \<lbrakk>\<Delta>\<rbrakk>"
  "\<Sigma>,\<Gamma> \<turnstile> \<theta> = [] \<in> \<lbrakk>\<Delta>\<rbrakk>"
  "\<Sigma>,\<Gamma> \<turnstile> \<theta> = \<sigma> \<in> \<lbrakk>(x,\<tau>)#\<Delta>\<rbrakk>"
  "\<Sigma>,\<Gamma> \<turnstile> (x,M)#\<theta> = \<sigma> \<in> \<lbrakk>\<Delta>\<rbrakk>"
  "\<Sigma>,\<Gamma> \<turnstile> \<theta> = (x,N)#\<sigma> \<in> \<lbrakk>\<Delta>\<rbrakk>"

lemma logrel_subst_valid:
  "\<Sigma>,\<Gamma> \<turnstile> \<theta> = \<sigma> \<in> \<lbrakk>\<Delta>\<rbrakk> \<Longrightarrow> valid_ctx \<Delta>"
  apply(induct rule:logrel_subst.inducts)
  apply(simp_all add:valid_ctx.intros)
  done

lemma logrel_weakening:
  "\<Gamma>' extends \<Gamma> ==> \<Sigma>,\<Gamma> \<turnstile> (M::hoas) = N \<in> \<lbrakk>\<tau>\<rbrakk> ==> \<Sigma>,\<Gamma>' \<turnstile> M = N \<in> \<lbrakk>\<tau>\<rbrakk>"
  apply(induct \<tau> arbitrary: \<Delta>' rule: hoas_ty.induct)
  apply(auto)
  apply(rule_tac x="Na" in exI)
  apply(auto  intro:  to_weakening)
  done

lemma logrel_subst_weakening:
  "\<Sigma>,\<Gamma> \<turnstile> (\<theta>::hoas_subst) = \<sigma> \<in> \<lbrakk>\<Delta>\<rbrakk> ==> \<Gamma>' extends \<Gamma> ==> \<Sigma>,\<Gamma>' \<turnstile> \<theta> = \<sigma> \<in> \<lbrakk>\<Delta>\<rbrakk>"
  apply(induct \<Sigma> \<Gamma> \<theta> \<sigma> \<Delta> rule:logrel_subst.induct) 
  apply(auto intro!:logrel_subst.intros intro:logrel_weakening)
  done

theorem logrel_main:
  shows "valid_ctx \<Gamma> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> M = M' \<in> \<lbrakk>\<tau>\<rbrakk> ==> \<exists> N. \<Sigma>,\<Gamma> \<turnstile> M \<Up> N : \<tau> \<and> \<Sigma>,\<Gamma> \<turnstile> M'\<Up> N : \<tau>"
  and "valid_ctx \<Gamma> \<Longrightarrow> \<exists> N. \<Sigma>,\<Gamma> \<turnstile> M \<down> N : \<tau> \<and> \<Sigma>,\<Gamma> \<turnstile> M' \<down> N : \<tau> ==> \<Sigma>,\<Gamma> \<turnstile> M = M' \<in> \<lbrakk>\<tau>\<rbrakk>"
  apply(induct \<tau> arbitrary: \<Gamma> M M' rule: hoas_ty.induct)
  apply(auto intro:to_can_atm)
    (* case 3 *)
  apply(rule_tac x="(\<Gamma>,M,M')" in  exists_fresh,finite_guess)
  apply(erule_tac x="(c,hoas_ty1)#\<Gamma>" in allE)
  apply(auto simp add:fresh_prod)
  apply(erule_tac x="VAR c" in allE)+
  apply(subgoal_tac "\<Sigma>,(c, hoas_ty1) # \<Gamma> \<turnstile> VAR c \<down> VAR c : hoas_ty1")
  prefer 2 apply(rule to_atm_var,simp)
  apply(subgoal_tac "\<Sigma>,(c, hoas_ty1) # \<Gamma> \<turnstile> VAR c = VAR c \<in> \<lbrakk>hoas_ty1\<rbrakk>")
  prefer 2 apply (blast)
  apply(subgoal_tac "\<Sigma>,(c, hoas_ty1) # \<Gamma> \<turnstile> APP M (VAR c) = APP M' (VAR c) \<in> \<lbrakk>hoas_ty2\<rbrakk>")
  prefer 2 apply(blast)
  apply(subgoal_tac "\<exists> K. \<Sigma>,(c, hoas_ty1) # \<Gamma> \<turnstile> APP M (VAR c) \<Up> K : hoas_ty2 \<and> \<Sigma>,(c, hoas_ty1) # \<Gamma> \<turnstile> APP M' (VAR c) \<Up> K : hoas_ty2")
  prefer 2 apply(blast)
  apply(clarsimp)
  apply(rule_tac x="LAM c K" in exI)
  apply(auto elim: to_can_fun simp add:fresh_prod)
    (* case 4 *)
  apply(subgoal_tac "\<exists> K. \<Sigma>,\<Delta>' \<turnstile> M1 \<Up> K :hoas_ty1 \<and> \<Sigma>,\<Delta>' \<turnstile> M1' \<Up> K :hoas_ty1")
  prefer 2 apply(blast)
  apply(clarify)
  apply(subgoal_tac "\<Sigma>,\<Delta>' \<turnstile> APP M M1 \<down> APP N K : hoas_ty2")
  prefer 2 apply(rule to_atm_app, erule to_weakening,simp,simp)
  apply(subgoal_tac "\<Sigma>,\<Delta>' \<turnstile> APP M' M1' \<down> APP N K : hoas_ty2")
  prefer 2 apply(rule to_atm_app, erule to_weakening,simp,simp)

  apply(blast)
  done
  
lemma logrel_symm:
  "\<Sigma>,\<Gamma> \<turnstile> (M::hoas) = N \<in> \<lbrakk>\<tau>\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> N = M \<in> \<lbrakk>\<tau>\<rbrakk>"
  apply(induct arbitrary: M M' \<Gamma> N rule:hoas_ty.inducts)
  by(auto)

lemma logrel_subst_symm:
  "\<Sigma>,\<Gamma> \<turnstile> (\<theta>::hoas_subst) = \<sigma> \<in> \<lbrakk>\<tau>\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> \<sigma> = \<theta> \<in> \<lbrakk>\<tau>\<rbrakk>"
  apply(induct rule:logrel_subst.inducts)
  by(auto intro:logrel_subst.intros logrel_symm)

lemma logrel_trans:
  "valid_sig \<Sigma> \<Longrightarrow> valid_ctx \<Gamma> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> (M::hoas) = N \<in> \<lbrakk>\<tau>\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> N = K \<in> \<lbrakk>\<tau>\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> M = K \<in> \<lbrakk>\<tau>\<rbrakk>"
  apply(induct arbitrary: \<Gamma> M N K rule:hoas_ty.inducts)
  (* base case *)
  apply(clarsimp)
  apply(subgoal_tac "Na = Nb")
  prefer 2 apply(erule to_determinacy,simp,simp,simp)
  apply(blast)
    (* fun ty case *) 
  apply(simp (no_asm_use),clarify)
  apply(subgoal_tac "\<Sigma>,\<Delta>' \<turnstile> M1 = M1 \<in> \<lbrakk>hoas_ty1\<rbrakk>")
  prefer 2 apply(blast intro:logrel_symm)
  apply(subgoal_tac " \<Sigma>,\<Delta>' \<turnstile> APP M M1 = APP N M1 \<in> \<lbrakk>hoas_ty2\<rbrakk>")
  prefer 2 apply(blast)
  apply(subgoal_tac " \<Sigma>,\<Delta>' \<turnstile> APP N M1 = APP K M1' \<in> \<lbrakk>hoas_ty2\<rbrakk>")
  prefer 2 apply(blast)
  apply(blast)
  done
  
lemma logrel_subst_trans:
  "valid_sig \<Sigma> \<Longrightarrow> valid_ctx \<Gamma> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> (\<theta>::hoas_subst) = \<sigma> \<in> \<lbrakk>\<Delta>\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> \<sigma> = \<sigma>' \<in> \<lbrakk>\<Delta>\<rbrakk> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> \<theta> = \<sigma>' \<in> \<lbrakk>\<Delta>\<rbrakk>"
  apply(induct \<Delta> arbitrary: \<Gamma> \<theta> \<sigma> \<sigma>')
  apply(auto elim:logrel_subst_inv )
  apply(erule logrel_subst_inv)+
  apply(clarify)
  apply(rule  logrel_subst_cons)
  apply(blast)
  apply(blast intro:logrel_trans)
  apply(assumption)
  done

lemma logrel_whr1: 
  "M \<leadsto> M' ==> \<Sigma>,\<Gamma> \<turnstile> M' = N \<in> \<lbrakk>\<tau>\<rbrakk> ==> \<Sigma>,\<Gamma> \<turnstile> M = N  \<in> \<lbrakk>\<tau>\<rbrakk>"
  apply(nominal_induct arbitrary: M M' \<Gamma> N rule:hoas_ty.strong_inducts)
  apply(clarsimp)
  apply(drule to_can_whr)
  apply(auto)
  apply(erule_tac x="\<Delta>'" in allE)
  apply(auto)
  apply(erule_tac x="M1" in allE)
  apply(blast)
  done

lemma logrel_whr2:
  "N \<leadsto> N' ==> \<Sigma>,\<Gamma> \<turnstile> M = N' \<in> \<lbrakk>\<tau>\<rbrakk> ==> \<Sigma>,\<Gamma> \<turnstile> M = N  \<in> \<lbrakk>\<tau>\<rbrakk>"
  by(auto dest: logrel_whr1  intro: logrel_symm)

lemma wf_var_logrel: 
  "\<lbrakk>\<Sigma>,\<Delta> \<turnstile> \<theta> = \<sigma> \<in> \<lbrakk>\<Gamma>\<rbrakk>;(x, \<tau>) \<in> set \<Gamma>\<rbrakk> \<Longrightarrow> \<Sigma>,\<Delta> \<turnstile> VAR x{\<theta>} = VAR x{\<sigma>} \<in> \<lbrakk>\<tau>\<rbrakk>"
  apply(induct \<Sigma> \<Delta> \<theta> \<sigma> \<Gamma> rule:logrel_subst.inducts)
  apply(auto simp add:fresh_not_in_ctx)
  done

lemma wf_logrel:
  "\<Sigma>,\<Gamma> \<turnstile> M : \<tau> \<Longrightarrow> valid_sig \<Sigma> \<Longrightarrow> valid_ctx \<Delta> \<Longrightarrow> \<Sigma>,\<Delta> \<turnstile> \<theta> = \<sigma> \<in> \<lbrakk>\<Gamma>\<rbrakk> \<Longrightarrow> \<Sigma>,\<Delta> \<turnstile> M{\<theta>} = M{\<sigma>} \<in> \<lbrakk>\<tau>\<rbrakk>"
  apply(nominal_induct \<Sigma> \<Gamma> M \<tau> avoiding: \<Delta> \<theta> \<sigma> rule: wf.strong_inducts)
(* case 1 *)
  apply(blast intro:wf_var_logrel)
   (* case 3: app = app *)
  apply(simp)
    (* case 2 *)
  apply(simp)
  apply(erule logrel_main)
  apply(blast intro:to_atm_cst) 
  (* case 4: lam=lam *)
  apply(clarsimp)
  apply(drule_tac x="\<Delta>'" in meta_spec,simp)
  apply(rule logrel_whr1)
  apply(rule whr_beta_strong)
  apply(rule logrel_whr2)
  apply(rule whr_beta_strong)
  apply(drule_tac x="(x,M1)#\<theta>" in meta_spec)
  apply(drule_tac x="(x,M1')#\<sigma>" in meta_spec)
  apply(subgoal_tac "sig,\<Delta>' \<turnstile> (x, M1) # \<theta> = (x,M1')#\<sigma> \<in> \<lbrakk>(x, ty1) # ctx\<rbrakk> ")
  prefer 2 apply(rule logrel_subst.intros)
           apply(erule logrel_subst_weakening)
	   apply(simp_all)
  apply(simp add: subst_fresh_merge)
  done

lemma logrel_beta:
  "\<lbrakk>\<Sigma>,\<Gamma> \<turnstile> APP (LAM x t) u : \<tau>; valid_sig \<Sigma>; valid_ctx \<Delta>; \<Sigma>,\<Delta> \<turnstile> \<theta> = \<sigma> \<in> \<lbrakk>\<Gamma>\<rbrakk>; x \<sharp> (\<Delta>,\<theta>,\<sigma>,\<Gamma>,u)\<rbrakk>
       \<Longrightarrow> \<Sigma>,\<Delta> \<turnstile> APP (LAM x t) u{\<theta>} = t{x::=u}{\<sigma>} \<in> \<lbrakk>\<tau>\<rbrakk>"
  apply(subgoal_tac "\<Sigma>,\<Gamma> \<turnstile> t{x::=u} : \<tau>")
  prefer 2 apply(rule whr_subject_reduction_ty)
  apply(rule whr_beta_strong,simp_all add:logrel_subst_valid)
  apply(drule wf_logrel,simp_all) back
  apply(rule logrel_whr1, rule whr_beta_strong)
  apply(simp add:fresh_prod subst_fresh_ssubst)  
  done

lemma logrel_eta:
  "\<lbrakk> \<Sigma>,\<Gamma> \<turnstile> t : \<tau> ~> \<tau>';valid_sig \<Sigma>;
        valid_ctx \<Delta>; x \<sharp> (\<Delta>,\<theta>,\<sigma>,\<Gamma>,t); \<Sigma>,\<Delta> \<turnstile> \<theta> = \<sigma> \<in> \<lbrakk>\<Gamma>\<rbrakk>\<rbrakk>
       \<Longrightarrow> \<Sigma>,\<Delta> \<turnstile> t{\<theta>} = LAM x (APP t (VAR x)){\<sigma>} \<in> \<lbrakk>\<tau> ~> \<tau>'\<rbrakk>"
  apply(auto simp add:lookup_fresh_id fresh_prod)
  apply(drule wf_logrel,simp_all)
  apply(rule logrel_whr2, rule whr_beta_strong)
  apply(simp)
  apply(subgoal_tac "t{\<sigma>}{x::=M1'} = t{\<sigma>}")
  prefer 2 apply(simp add:subst_fresh_id ssubst_fresh fresh_prod) 
  apply(auto)
  done

lemma wf_equiv_subst_logrel:
  "\<Sigma>,\<Gamma> \<turnstile> M \<approx> N : \<tau> \<Longrightarrow> valid_sig \<Sigma> \<Longrightarrow> valid_ctx \<Delta> \<Longrightarrow> \<Sigma>,\<Delta> \<turnstile> \<theta> = \<sigma> \<in> \<lbrakk>\<Gamma>\<rbrakk> \<Longrightarrow> \<Sigma>,\<Delta> \<turnstile> M{\<theta>} = N{\<sigma>} \<in> \<lbrakk>\<tau>\<rbrakk>"
  apply(nominal_induct \<Sigma> \<Gamma> M N \<tau> avoiding: \<Delta> \<theta> \<sigma> rule:wf_equiv.strong_inducts)
    (* case 1: var=var *)
  apply(blast intro:wf_var_logrel)
    (* case 2: cst=cst *)
  apply(simp)
  apply(erule logrel_main)
  apply(blast intro:to_atm_cst) 
    (* case 3: app = app *)
  apply(clarsimp)
  (* case 4: lam=lam *)
  apply(clarsimp)
  apply(drule_tac x="\<Delta>'" in meta_spec,simp)
  apply(rule logrel_whr1)
  apply(rule whr_beta_strong)
  apply(rule logrel_whr2)
  apply(rule whr_beta_strong)
  apply(drule_tac x="(x,M1)#\<theta>" in meta_spec)
  apply(drule_tac x="(x,M1')#\<sigma>" in meta_spec)
  apply(subgoal_tac "\<Sigma>,\<Delta>' \<turnstile> (x, M1) # \<theta> = (x,M1')#\<sigma> \<in> \<lbrakk>(x, \<tau>) # \<Gamma>\<rbrakk> ")
  prefer 2 apply(rule logrel_subst.intros)
           apply(erule logrel_subst_weakening)
	   apply(simp,simp,simp)
  apply(simp add: subst_fresh_merge)
    (* case 5: beta *)
  apply(erule logrel_beta, simp,simp,simp, simp add:fresh_prod)
(* case 6: eta *)
  apply(erule logrel_eta,simp,simp,simp,simp add:fresh_prod)
  (* case 7: symmetry *)
  apply(drule logrel_subst_symm, rule logrel_symm)
  apply(blast)
    (* case 8: transitivity *)
  apply(subgoal_tac " \<Sigma>,\<Delta> \<turnstile> \<theta> = \<theta> \<in> \<lbrakk>\<Gamma>\<rbrakk>")
  prefer 2 apply(blast intro:logrel_subst_trans logrel_subst_symm) 
  apply(subgoal_tac " \<Sigma>,\<Delta> \<turnstile> t{\<theta>} = u{\<theta>} \<in> \<lbrakk>\<tau>\<rbrakk>")
  prefer 2 apply(blast)
  apply(blast intro:logrel_trans)
  done

 
lemma id_subst_logrel:
  "valid_ctx \<Gamma> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> id_sub \<Gamma> = id_sub \<Gamma>  \<in> \<lbrakk>\<Gamma>\<rbrakk>"
  apply(induct \<Gamma>)
  apply(simp,blast intro:logrel_subst_nil)
  apply(erule valid_ctx.cases, simp_all,clarsimp)
  apply(rule logrel_subst_cons)
  apply(erule logrel_subst_weakening)
  apply(auto)
  apply(rule logrel_main,auto)
  apply(rule_tac x="VAR x" in exI,auto intro: to_atm_var)
  done


lemma wf_implies_lr:
  "valid_ctx \<Gamma> \<Longrightarrow> valid_sig \<Sigma> \<Longrightarrow> \<Sigma>, \<Gamma> \<turnstile> M \<approx> N : \<tau> \<Longrightarrow> \<Sigma>,\<Gamma> \<turnstile> M = N \<in>  \<lbrakk>\<tau>\<rbrakk>"
  apply(drule wf_equiv_subst_logrel,simp)
  apply(simp)
  apply(erule id_subst_logrel)
  apply(simp add: id_subst_id)
  done

theorem completeness:
  "\<Sigma>, \<Gamma> \<turnstile> M \<approx> M' : \<tau> \<Longrightarrow> valid_ctx \<Gamma> \<Longrightarrow> valid_sig \<Sigma> \<Longrightarrow> \<exists> N. \<Sigma>,\<Gamma> \<turnstile> M \<Up> N : \<tau> \<and> \<Sigma>,\<Gamma> \<turnstile> M' \<Up> N : \<tau>"
  apply(frule wf_implies_lr,simp_all)
  apply(rule logrel_main,simp_all)
  done

theorem uniqueness:
  "\<Sigma>, \<Gamma> \<turnstile> M \<approx> M' : \<tau> \<Longrightarrow> valid_ctx \<Gamma> \<Longrightarrow> valid_sig \<Sigma> \<Longrightarrow> \<exists>! N. \<Sigma>,\<Gamma> \<turnstile> M \<Up> N : \<tau> \<and> \<Sigma>,\<Gamma> \<turnstile> M' \<Up> N : \<tau>"
  apply(frule completeness,simp_all)
  apply(auto)  
  apply(blast intro: to_determinacy)
  done

theorem canonicalization:
  "\<Sigma>, \<Gamma> \<turnstile> M : \<tau> \<Longrightarrow> valid_ctx \<Gamma> \<Longrightarrow> valid_sig \<Sigma> \<Longrightarrow> \<exists>! N. \<Sigma>,\<Gamma> \<turnstile> M \<Up> N : \<tau>"
  apply(auto)
  apply(frule wf_equiv_refl)
  apply(frule completeness,auto)
  apply(blast intro: to_determinacy)
  done



end