theory Lambda 
imports Nominal Name

begin

nominal_datatype lam =
    Var "var"
  | App "lam" "lam"
  | Lam "\<guillemotleft>var\<guillemotright>lam"


nominal_primrec 
subst :: "lam \<Rightarrow> var \<Rightarrow> lam \<Rightarrow> lam" ("_[_::=_]" [100,100,100] 100)
where
  "(Var x)[y::=t] = (if x = y then t else (Var x))"
| "(App t1 t2)[y::=t] = App (t1[y::=t]) (t2[y::=t])"
| "x\<sharp>(y,t) \<Longrightarrow> (Lam x t0)[y::=t] = Lam x (t0[y::=t])"
  apply(finite_guess+)
  apply(rule TrueI)+
  apply(simp add:abs_fresh)
  apply(fresh_guess)+
  done

lemma subst_eqvt[eqvt]:
  fixes pi:: "var prm"
  and t1 t2::"lam"
  shows "pi\<bullet>(t1[b::=t2]) = (pi\<bullet>t1)[(pi\<bullet>b)::=(pi\<bullet>t2)]"
  apply(nominal_induct t1 avoiding: b t2 rule: lam.strong_induct)
  apply(auto simp add: perm_bij fresh_prod fresh_atm fresh_bij)
  done

lemma subst_supp: 
  shows "supp(t1[a::=t2]) \<subseteq> (((supp(t1)-{a})\<union>supp(t2))::var set)"
apply(nominal_induct t1 avoiding: a t2 rule: lam.strong_induct)
apply(auto simp add: lam.supp supp_atm fresh_prod abs_supp)
apply(blast)+
done

lemma subst_fresh : "x \<sharp> n \<Longrightarrow> x \<sharp> m[x::=n]"
  apply(nominal_induct m avoiding: x n rule:lam.strong_induct)
  apply(auto simp add:lam.fresh fresh_atm abs_fresh )
  done



lemma subst_swap: "\<lbrakk>y \<sharp> t;y \<sharp> u\<rbrakk> \<Longrightarrow> ([(y,x)]\<bullet> t)[y::=u] = t[x::=u]"
  apply(nominal_induct t avoiding: x y u rule: lam.strong_inducts)
  apply(auto simp add: lam.inject calc_atm fresh_atm abs_fresh) 
  done


text {* Well-formedness (mentions only names in given set *}

inductive wf :: "var list \<Rightarrow> lam \<Rightarrow> bool" ("_ \<turnstile> _" [60,60] 60)
where
  wf_var[intro!] : "x \<in> set X \<Longrightarrow> X \<turnstile> Var x"
| wf_app[intro!] : "\<lbrakk>X \<turnstile> t1;X \<turnstile> t2\<rbrakk> \<Longrightarrow> X \<turnstile> App t1 t2"
| wf_lam[intro!] : "\<lbrakk>x \<sharp> X; x#X \<turnstile> t1\<rbrakk> \<Longrightarrow> X \<turnstile> Lam x t1"

inductive_cases wf_inv[elim!]: 
  "X \<turnstile> Var x"
  "X \<turnstile> App t1 t2"
  "X \<turnstile> Lam x t1"

equivariance wf

nominal_inductive wf
  apply(clarsimp)
  apply(simp add:abs_fresh)+
  done



lemma fresh_not_in: 
  fixes x::var
  and X::"var list"
  shows "x \<sharp> X \<Longrightarrow> ((x) \<notin> set X)"
  apply(induct X)
  by(auto simp add:fresh_prod fresh_list_cons fresh_atm)

lemma wf_fresh:
  fixes x::var
  shows "\<lbrakk>X \<turnstile> t; x \<sharp> X\<rbrakk> \<Longrightarrow> x \<sharp> t" 
  apply(nominal_induct X t avoiding: x rule:Lambda.wf.strong_inducts)
  apply(auto dest: fresh_not_in simp add: abs_fresh fresh_list_cons fresh_atm)
  done

lemma wf_weak:
  "\<lbrakk>X \<turnstile> t; set X \<subseteq> set X'\<rbrakk> \<Longrightarrow> X' \<turnstile> t"
  apply(nominal_induct avoiding: X' rule:wf.strong_inducts)
  apply(auto)
  apply(subgoal_tac "set (x#X)  \<subseteq> set (x#X')", auto)
  done

lemma wf_subst_general:
  "X'@[x]@X \<turnstile> t \<Longrightarrow> X \<turnstile> u \<Longrightarrow> X'@X \<turnstile> t[x::=u]"
  apply(nominal_induct X\<equiv>"X'@[x]@X" t avoiding: X x X' u rule: wf.strong_inducts)
  apply(auto intro:wf_weak simp add:fresh_list_append)
  apply(drule_tac x="Xa" in meta_spec)
  apply(drule_tac x="xa" in meta_spec)
  apply(drule_tac x="x#X'" in meta_spec)
  apply(drule_tac x="u" in meta_spec)
  apply(drule meta_mp,auto simp add:fresh_list_append)
  done

lemma wf_subst:
  "x#X \<turnstile> t \<Longrightarrow> X \<turnstile> u \<Longrightarrow> X \<turnstile> t[x::=u]"
  apply(subgoal_tac "[]@X \<turnstile> t[x::=u]",simp)
  apply(rule_tac X'="[]" in wf_subst_general,auto)
  done


inductive valid :: "var list \<Rightarrow> bool"
where
  valid_emp[simp] : "valid []"
| valid_cons[intro!] : "\<lbrakk>x \<sharp> X; valid X\<rbrakk> \<Longrightarrow> valid (x#X)"




text {* Abbreviations for the sets of well-formed terms *}

constdefs LAMBDA_in :: "var list \<Rightarrow> lam set" ("LAMBDA[_]" [100] 100)
"LAMBDA[X] == {t. X \<turnstile> t}"

constdefs LAMBDA :: "lam set"
"LAMBDA == UN X : {Y. valid Y}.  (LAMBDA[X])"

inductive red :: "lam \<Rightarrow> lam \<Rightarrow> bool" ("_ \<mapsto> _" [100,100] 100)
where red_beta[intro] : 
  "x \<sharp> n \<Longrightarrow> App (Lam x m) n \<mapsto> m[x::=n]"
| red_app1[intro] : 
  "m \<mapsto> m' \<Longrightarrow> App m n \<mapsto> App m' n"
| red_app2[intro] : 
  "n \<mapsto>n' \<Longrightarrow> App m n \<mapsto> App m n'"
| red_lam[intro!]:
  "m\<mapsto> m' ==> Lam x m \<mapsto> Lam x m'"

equivariance red

nominal_inductive red
  apply(clarsimp)
  apply(simp add:abs_fresh subst_fresh)+
  done


lemma red_beta_strong: "App (Lam x t) u \<mapsto> t[x::=u]"
proof -
  obtain y::"var" where fc: "y\<sharp>(x,t,u)" by (rule_tac exists_fresh) (auto simp add: fs_var1)
  have "App (Lam x t) u = App (Lam y ([(y,x)]\<bullet> t)) u" using fc 
    by (auto simp add: lam.inject alpha' fresh_atm fresh_prod)
  also have "\<dots> \<mapsto>  ([(y,x)]\<bullet> t)[y::=u]" using fc by (rule_tac Lambda.red_beta) (simp)
  also have "\<dots> = t[x::=u]" using fc by (simp add: subst_swap)
  finally show "App (Lam x t) u \<mapsto> t[x::=u]" by simp
qed

end
